// -*- c++ -*-
// gridagent.h
// Douglas S. J. De Couto
// 28 October 2000

// $Id: gridagent.h,v 1.12 2001/01/07 22:12:31 decouto Exp $

#ifndef _GridAgent_h
#define _GridAgent_h

#include <agent.h>
#include <packet.h>
#include <trace.h>
#include <cmu/priqueue.h>
#include <cmu/mac.h>
#include <cmu/mac-802_11.h>
#include "hdr_grid.h"
#include "ipaddress.h"
#include "vector.hh"

class GridAgent;
class GridRT;

/* Timer helpers */

class GridHelloTimer : public TimerHandler {
public:

  static const double DEF_HELLO_PERIOD = 2; // secs
  // some positive fraction of the period is added to itself period, up to the specified jitter
  static const double HELLO_JITTER = 0.3; 

  GridHelloTimer(GridAgent *a, double period = DEF_HELLO_PERIOD) 
    : TimerHandler(), agent_(a), t_(period) { }
  void expire(Event *);
private:
  GridAgent *agent_;
  double t_;
  friend class GridAgent;
};

class GridExpireTimer : public TimerHandler {
public:

  static const double DEF_EXPIRE_PERIOD = 6; // secs
  // some positive fraction of the period is added to itself period, up to the specified jitter
  static const double EXPIRE_JITTER = 0.3; 

  GridExpireTimer(GridAgent *a, double period = DEF_EXPIRE_PERIOD) 
    : TimerHandler(), agent_(a), t_(period) { }
  void expire(Event *);
private:
  GridAgent *agent_;
  double t_;
  friend class GridAgent;
};

class GridTriggerTimer : public TimerHandler {
public:

  static const double DEF_TRIGGER_PERIOD = 1; 
  static const double TRIGGER_JITTER = 0.3; 

  GridTriggerTimer(GridAgent *a, double period = DEF_TRIGGER_PERIOD) 
    : TimerHandler(), agent_(a), t_(period) { }
  void expire(Event *);
private:
  GridAgent *agent_;
  double t_;
  friend class GridAgent;
};

class GridAdvertiseLocTimer : public TimerHandler {
public:

  static const double DEF_ADVERTISE_PERIOD = 0.25; 
  static const double ADVERTISE_JITTER = 0.3; 

  GridAdvertiseLocTimer(GridAgent *a, double period = DEF_ADVERTISE_PERIOD) 
    : TimerHandler(), agent_(a), t_(period) { }
  void expire(Event *);
private:
  GridAgent *agent_;
  double t_;
  friend class GridAgent;
};

/* End timer helpers */


class GridAgent : public Agent {
public:
  virtual int command(int argc, const char* const *argv);
  virtual void recv(Packet *, Handler *callback = 0);

  GridAgent();
  ~GridAgent();

  virtual void Terminate();

private:
  void start();
  void trace(char* fmt, ...);

  void originate_packet(Packet *p);
  void receive_packet(Packet *p);
  void handle_encap_packet(Packet *p);
  void forward_packet(Packet *p, bool is_rescue = false);

  void set_tx_loc(grid_hdr *gh);
  void set_src_loc(grid_hdr *gh);

  struct grid_location get_loc(); // node's current location
  struct grid_location get_loc(int ip); // any node's current loc, use the ns host byte order address

  bool get_advertised_loc(grid_location *loc, double *when); // false iff no loc advertised
  bool get_advertised_loc(grid_location *loc, double *when, int ip); 

  void advertise_loc(grid_location loc);
  void withdraw_loc();

  Packet *alloc_pkt(int sz);

  Vector<Packet *> extract_pkts_from_ifq(IPAddress next_hop);

  void expire_timer(); // expire route table entries
  void hello_timer(); // broadcast new hello msg
  void trigger_timer(); // broadcast triggered route entries
  void advertise_loc_timer(); // ``advertise'' current location

  static void tx_failed_callback(Packet *p, void *thunk);
  void handle_tx_failed(Packet *p);

  void print_table_sizes();

  nsaddr_t ip_; 
  // note, this is just some int in host byte order; all the packet
  // formats, and the GridRT use network byte order ip address.  must
  // be careful to swap when putting the agent's ip_ field into grid
  // packets, or passing to the GridRT object.
  nsaddr_t eth_;  

  Mac802_11 *mac_; // mac layer, we won't use the ns link layer
  PriQueue *ifq_; // interface queue we send packets to, may be a log target
  PriQueue *real_ifq_; // the *actual* interface queue, we can pull packts out of it, etc.
  Trace *logtarget_; // logger
  MobileNode *mobile_node_; // the moving node this agent associates with
  
  int off_mac_;
  int off_ll_;
  int off_ip_;
  int off_grid_lr_;
  int off_grid_encap_;
  int off_grid_proxy_;

  friend class GridHelloTimer;
  friend class GridExpireTimer;
  friend class GridTriggerTimer;
  friend class GridAdvertiseLocTimer;
  friend class GridRT;

  GridHelloTimer hello_timer_;
  GridExpireTimer expire_timer_;
  GridTriggerTimer trigger_timer_;
  GridAdvertiseLocTimer advertise_loc_timer_;

  GridRT *route_table_;
  int max_hello_sz_;
  int use_geo_fwd_; // 0 means no, 1 means yes, other values for future use

  int max_hops_;
  int propagate_locs_;
  int num_good_locs_;

  bool proxy_straw_man_;

  IPAddress proxy_ip_;
  bool have_proxy_loc_; // know_loc_ ==> have_proxy_loc_ is false.  the opposite is not true.
  bool know_loc_;

  int count(Vector<int> v, int val);

  static int max_rt_sz_;
  static int max_geo_rt_sz_;
  static int max_ignorant_rt_sz_;
};



#endif


