// gridrt.cc
// Douglas S. J. De Couto
// 16 November 2000

// $Id: gridrt.cc,v 1.13 2001/01/08 20:06:32 decouto Exp $

#include <sys/types.h>
#include <netinet/in.h>

#include <scheduler.h>
#include <cmu/mac.h>
#include "gridrt.h"
#include "gridagent.h"

#include "vector.cc"
#include "bighashmap.cc"

/* NOTE ON TIMEOUTS: there are two different timeouts.  The `age' of a
   routing entry is approximately how much time has passed since the
   entry was originally produced by the node it points too.  The age
   is increased everytime an entry is propagated, as well as with the
   passage of time, to prevent stale data from sticking around
   (actually, the age is set, and then decremented, the entry is
   thrown away when its age is ~0). 

   The ``jiffies' of an entry are how long that entry has been in this
   particular routing table, we discard entries that are too old here
   also.  It could be that the age makes jiffies redundant. */

#if 0
struct has_ip : public unary_function<grid_nbr_entry, bool> { 
  unsigned int _ip;
  has_ip(unsigned int ip) : _ip(ip) { }
  bool operator()(const grid_nbr_entry &n) const { return n.ip == _ip; }
};

struct has_ip_far : public unary_function<GridRT::far_entry, bool> { 
  unsigned int _ip;
  has_ip_far(unsigned int ip) : _ip(ip) { }
  bool operator()(const GridRT::far_entry &n) const { return n.nbr.ip == _ip; }
};
#endif

void
GridRT::run_expire_hook() 
{
  typedef Vector<grid_nbr_entry> x_t;
  x_t expired_info = expire_routes();
  
  // update the loc proxy info, we may no longer have route to an old proxy
  if (do_proxy_stuff())
    pick_new_loc_proxy();

  int num_to_send = min(expired_info.size(), _max_hello_entries);
  
  if (num_to_send > 0) {
    // make and send the packet advertising any broken routes
    x_t update_vec(expired_info.sub(0, num_to_send));
    send_routing_update(update_vec);
  }
  _leftover_expired = expired_info.sub(num_to_send, expired_info.size());
}

void
GridRT::pick_new_loc_proxy()
{
  assert(do_proxy_stuff());

  // pick a node which knows its location as our proxy.  pick the one
  // that is the least hops away.
  bool first_time = true;
  have_proxy() = false;
  for (FarTable::Iterator iter = _rtes.first(); iter; iter++) {
    if (!iter.value().nbr.loc_good)
      continue;
    if (first_time || iter.value().nbr.num_hops < rt_get_ref(proxy_ip()).nbr.num_hops) {
      proxy_ip() = iter.key();
      have_proxy() = true;
    } 
    first_time = false;
  }

  if (have_proxy())
    assert(rt_contains(proxy_ip()));
}


Vector<grid_nbr_entry>
GridRT::expire_routes()
{
  /* at the end of this function, the _leftover_expired vector is empty,
     and all advertisement-pending expired routes are returned */

  double jiff = Scheduler::instance().clock();
  
  // XXX not sure if we are allowed to iterate while modifying map
  // (i.e. erasing entries), so figure out what to expire first.
  typedef BigHashMap<IPAddress, bool> xa_t;
  xa_t expired_addresses;
  typedef Vector<grid_nbr_entry> vgne_t;
  vgne_t expired_nbrs;
  Vector<char *> why;

  // find the expired immediate entries
  for (Table::Iterator iter = _addresses.first(); iter; iter++) 
    if (jiff - iter.value().last_updated_jiffies > _timeout_jiffies)
      expired_addresses.insert(iter.key(), true);

  // find expired routing entries -- either we have had the entry for
  // too long, or it has been flying around the whole network for too
  // long, or we have expired the next hop from our immediate neighbor
  const double epsilon = 0.0001;
  for (FarTable::Iterator iter = _rtes.first(); iter; iter++) {
    assert(iter.value().nbr.num_hops > 0);
    double how_old = jiff - iter.value().last_updated_jiffies;
    if (how_old > _timeout_jiffies ||
	decr_age(iter.value().nbr.age, how_old) < epsilon ||
	expired_addresses.findp(iter.value().nbr.next_hop_ip) != 0) {
      grid_nbr_entry nbr = iter.value().nbr;
      nbr.num_hops = 0;
      nbr.seq_no++; // odd numbers indicate broken routes
      nbr.age = grid_hello::MAX_AGE_DEFAULT;
      expired_nbrs.push_back(nbr);

      if (how_old > _timeout_jiffies)
	why.push_back("bad-jiff");
      else if (decr_age(iter.value().nbr.age, how_old) < epsilon)
	why.push_back("bad-age");
      else if (expired_addresses.findp(iter.value().nbr.next_hop_ip) != 0)
	why.push_back("expired-next-hop");
      else 
	why.push_back("uknown-expiry-reason");
    }
  }

  // remove expired immediate nbr entries
  for (xa_t::Iterator iter = expired_addresses.first(); iter; iter++) {
    if (_verbose)
      _agent->trace("S %.9f _%d_ expiring address for %s", jiff, _agent->ip_, iter.key().s().cc());
    assert(_addresses.remove(iter.key()));
  }

  // remove expired route table entry
  for (int i = 0; i < expired_nbrs.size(); i++) {
    if (_verbose) 
      _agent->trace("S %.9f _%d_ expiring route for %s (%s)", jiff, _agent->ip_, IPAddress(expired_nbrs[i].ip).s().cc(), why[i]);
    rt_erase(expired_nbrs[i].ip);
  }


  // copy leftover expired neighbors into the return vector, but don't
  // replace any newly expired entries with old ones for the same
  // address (so we don't screw up sequence numbers)
  for (int i = 0; i < _leftover_expired.size(); i++) {
    int i2 = find_ip(expired_nbrs, _leftover_expired[i].ip);
    if (i2 < 0)
      expired_nbrs.push_back(_leftover_expired[i]);
    else if (_leftover_expired[i].seq_no > expired_nbrs[i2].seq_no)
      expired_nbrs[i2] = _leftover_expired[i]; // somehow we got an older entry which we ended up expiring
  }

  _leftover_expired.clear();
  
  return expired_nbrs;
}


void
GridRT::run_trigger_hook()
{
  double jiff = Scheduler::instance().clock();

  /* we may not get to advertise all the triggered ``good'' routes and
     broken routes.  We will give priority to the new ``good'' routes,
     and then put in as many broken routes as we can.  Any leftover
     route ads will be left for next time */

  sort(_triggered_good_routes);
  sort(_triggered_broken_routes);

  size_t num_to_send = min(_triggered_good_routes.size() + _triggered_broken_routes.size(), 
			   _max_hello_entries);
  size_t num_sent = 0;
  Vector<grid_nbr_entry> update_vec;
  // add any good updates
  int i = 0;
  for ( ; i < _triggered_good_routes.size() && num_sent < num_to_send; i++) {
    far_entry &f = _triggered_good_routes[i];
    if (f.nbr.age - (jiff - f.when_triggered) > 0) {
      f.nbr.age -= (jiff - f.when_triggered);
      update_vec.push_back(f.nbr);
      num_sent++;
    }
  }
  // throw away sent triggered updates, keep all the rest --- note we
  // throw out any outdated triggered routes, even if they weren't
  // sent
  _triggered_good_routes = _triggered_good_routes.sub(i, _triggered_good_routes.size());
  // add any broken route updates
  for (i = 0; i < _triggered_broken_routes.size() && num_sent < num_to_send; i++) {
    far_entry &f = _triggered_broken_routes[i];
    if (f.nbr.age - (jiff - f.when_triggered) > 0) {
      f.nbr.age -= (jiff - f.when_triggered);
      update_vec.push_back(f.nbr);
      num_sent++;
    }
  }
  // throw away sent broken routes
  _triggered_broken_routes = _triggered_broken_routes.sub(i, _triggered_broken_routes.size());

  send_routing_update(update_vec);
}

void
GridRT::run_hello_hook()
{
  double jiff = Scheduler::instance().clock();

  /* XXX is this a bug?  we expire some routes, but don't advertise
     them as broken anymore... */
  Vector<grid_nbr_entry> expired_routes = expire_routes();

  // every entry ought to be advertised at least this often
  double max_ad_period = _rtes.size() * _hello_jiffies / _max_hello_entries;

  Vector<far_entry> recently_sent;
  Vector<far_entry> not_recently_sent;
  Vector<far_entry> never_sent;
  Vector<far_entry> loc_routes;
  for (FarTable::Iterator iter = _rtes.first(); iter; iter++) {
    /* XXX if everyone is using the same max-hops parameter, we could
       leave out all of our entries that are exactly max-hops hops
       away, because we know those entries will be greater than
       max-hops at any neighbor.  but, let's leave it in case we have
       different max-hops across the network */
    /* because we called expire_routes() at the top of this function,
       we know we are not propagating any route entries with age of 0
       or that have timed out */

    far_entry fe = iter.value();
    if (do_proxy_stuff1() && iter.value().nbr.loc_good) {
      // if we don't have enough routes to nodes with locations, just
      // push back any route with a location
      if (loc_routes.size() < _agent->num_good_locs_) {
	loc_routes.push_back(iter.value());
	continue;
      }
      else {
	// we already have enough routes with a location; kick out the
	// entry with the most hops that is more hops than fe (if
	// any), and deal with it below
	for (int i = 0; i < loc_routes.size(); i++) {
	  if (fe.nbr.num_hops < loc_routes[i].nbr.num_hops) {
	    far_entry t = fe;
	    fe = loc_routes[i]; 
	    loc_routes[i] = t;
	  }
	}
      }
    }
    else if (do_proxy_stuff2() && iter.value().nbr.loc_good) {
      // use a different criteria for choosing loc aware routes to
      // advertise, e.g. cycle fairly through all that we know, don't
      // just pick the 2 closest.
    }

    if (fe.nbr.num_hops > _max_hops)
      continue; // don't advertise it... special loc enries are taken care of above

    if (fe.last_advertised < 0)
      never_sent.push_back(fe);
    else if (jiff - fe.last_advertised < max_ad_period)
      recently_sent.push_back(fe);
    else 
      not_recently_sent.push_back(fe);
  }

  // sort by last_advertised field
  sort(recently_sent);
  sort(not_recently_sent);
  
  Vector<grid_nbr_entry> rte_entries;

  int num_to_send = 0;

  // if running location propagation function, add entries with good
  // locations
  if (do_proxy_stuff()) {
    for (int i = 0; num_to_send < _max_hello_entries && i < loc_routes.size(); i++, num_to_send++) {
      rte_entries.push_back(loc_routes[i].nbr);
      rt_get_ref(loc_routes[i].nbr.ip).last_advertised = jiff;
    }
  }

  // add any entries that have never been sent
  for (int i = 0; num_to_send < _max_hello_entries && i < never_sent.size(); i++, num_to_send++) {
    rte_entries.push_back(never_sent[i].nbr);
    rt_get_ref(never_sent[i].nbr.ip).last_advertised = jiff;
  }

  // add any entries that have been waiting too long to be sent
  for (int i = 0; num_to_send < _max_hello_entries && i < not_recently_sent.size(); i++, num_to_send++) {
    rte_entries.push_back(not_recently_sent[i].nbr);
    rt_get_ref(not_recently_sent[i].nbr.ip).last_advertised = jiff;
  }

  // add any other entries
  for (int i = 0; num_to_send < _max_hello_entries && i < recently_sent.size(); i++, num_to_send++) {
    rte_entries.push_back(recently_sent[i].nbr);
    rt_get_ref(recently_sent[i].nbr.ip).last_advertised = jiff;
  }

  // add any expired routes we have room for
  int i2;
  for (i2 = 0; num_to_send < _max_hello_entries && i2 < expired_routes.size(); i2++, num_to_send++) {
    rte_entries.push_back(expired_routes[i2]);
  }

  // remember any unadvertised broken routes
  _leftover_expired = expired_routes.sub(i2, expired_routes.size());

  // need good proxy before we can send the rte update
  if (do_proxy_stuff())
    pick_new_loc_proxy();

  // make and send the packet
  send_routing_update(rte_entries);
  // Update the sequence number for periodic updates (but not for
  // triggered updates).  Has to happen after sending so that the
  // hello uses the current seq_no.
  /* originating sequence numbers are even, starting at 0.  odd
     numbers are reserved for other nodes to advertise a broken route
     to us.  from DSDV paper. */
  _seq_no += 2;

  _triggered_good_routes.clear();
}


void
GridRT::run_proxy_hook()
{
  
}

void
GridRT::process_proxy_updates(Packet *p, double jiff)
{

}

void
GridRT::process_route_updates(Packet *p, double jiff)
{
  struct hdr_grid_lr *lr_header = (hdr_grid_lr *) p->access(_agent->off_grid_lr_);
  void *route_data = p->accessdata();
  
  struct grid_hdr *gh = &lr_header->gh_;
  struct grid_hello *hlo = &lr_header->hello_;
  
  Vector<IPAddress> new_rtes;

  /*
   * update far nbr info with this hello sender info -- list sender as
   * its own next hop.  */
  if (!rt_contains(gh->tx_ip))
    new_rtes.push_back(gh->tx_ip);
    
  far_entry fe = far_entry(jiff, grid_nbr_entry(gh->tx_ip, gh->tx_ip, 1, hlo->seq_no));
  fe.nbr.loc = gh->tx_loc;
  fe.nbr.loc_err = gh->tx_loc_err;
  fe.nbr.loc_good = gh->tx_loc_good;
  fe.nbr.age = decr_age(hlo->age, grid_hello::MIN_AGE_DECREMENT);
  if (hlo->max_hops > 0) 
    fe.nbr.max_hops = hlo->max_hops - 1;
  
  rt_insert(fe);

  if (_verbose)
    _agent->trace("S %.9f _%d_ adding 1-hop route for %s", jiff, _agent->ip_, IPAddress(gh->tx_ip).s().cc());
  

  /* 
   * add this sender's nbrs to our far neighbor list.  
   */
  Vector<far_entry> triggered_rtes;
  Vector<far_entry> broken_rtes;
  // rtes_to_delete will have a entry for each element of broken_rtes,
  // plus entries for broken route advertisements that are too old to
  // propagate, but young enough to believe ourselves.
  Vector<IPAddress> rtes_to_delete;
  
  // loop through all the route entries in the packet
  for (int i = 0; i < hlo->num_nbrs; i++) {
    grid_nbr_entry *curr = (grid_nbr_entry *) ((char *) route_data + i*sizeof(grid_nbr_entry));

    if (curr->ip == (unsigned int) _ipaddr)
      continue; // we already know how to get to ourself -- don't want to advertise some other strange route to us!

    if (curr->next_hop_ip == (unsigned int) _ipaddr)
      continue; // pseduo-split-horizon: ignore routes from nbrs that go back through us

    if (curr->num_hops == 0) {
      /* this entry indicates a broken route.  if the seq_no is
	 newer than any information we have, AND we route to the
	 specified address using this packet's sender as next hop,
	 remove the broken route.  propagate broken route info.
	 if the seq_no is older than some good route information
	 we have, advertise our new information to overthrow the
	 old broken route info we received */

      IPAddress broken_ip(curr->ip);
      
      if (rt_contains(broken_ip)) {
	far_entry fe = rt_get_ref(broken_ip);

	if (curr->seq_no > fe.nbr.seq_no && 
	    fe.nbr.next_hop_ip == gh->ip) {
	  // invalidate a route we have through this next hop
	  far_entry new_entry(fe);
	  new_entry.nbr.num_hops = 0;
	  // broken route info should be odd seq_no's
	  new_entry.nbr.seq_no = curr->seq_no;
	  assert((new_entry.nbr.seq_no & 1) == 1); 
	  new_entry.nbr.age = decr_age(curr->age, grid_hello::MIN_AGE_DECREMENT);
	  new_entry.when_triggered = jiff;
	  if (new_entry.nbr.age > 0) // don't propagate old messages
	    broken_rtes.push_back(new_entry);
	  rtes_to_delete.push_back(broken_ip);
	  if (_verbose)
	    _agent->trace("S %.9f _%d_ removing broken route to %s", jiff, _agent->ip_, broken_ip.s().cc());

	}
	else if (curr->seq_no < fe.nbr.seq_no) {
	  // we know more recent info about a route that this
	  // entry is trying to invalidate
	  far_entry new_entry(fe);
	  assert((new_entry.nbr.seq_no & 1) == 0);
	  new_entry.nbr.age = decr_age(new_entry.nbr.age, jiff);
	  new_entry.when_triggered = jiff;
	  if (new_entry.nbr.age > 0) {
	    triggered_rtes.push_back(new_entry);
	    if (_verbose)
	      _agent->trace("S %.9f _%d_ updating broken route to %s with new route", jiff, _agent->ip_, broken_ip.s().cc());
	  }
	}
      }
      else
	; // else we never had a route to this broken destination anyway

      continue; // process next route in message
    }

    // XXX bug bug bug??? perhaps the 2nd condition shouldn't have the do_proxy_stuff() term?

    if (!(curr->num_hops + 1 <= _max_hops                  // the dest is within out routing radius
	  || (/* do_proxy_stuff() && */ curr->max_hops > 0)// or, the dest specified a routing radius
	  || (do_proxy_stuff() && curr->loc_good)          // or, this is a route to node with known loc
      	  || (do_proxy_stuff_with_loc() && curr->loc_good))) { // or, remember this route for connectedness
      // skip this one, we don't care about nbrs too many hops away,
      // unless: they have a max_hops entry that says we should keep
      //  propagating the entry,
      // or: they know their location and we don't, 
      // or: they know their location and we do...(but don't advertise
      //  it again) -- i.e. even if know own location, remember any
      //  route we hear to someone with location --- want to maintain
      //  connectedness of location-aware nodes
      continue; 
    }

    unsigned int curr_seq = curr->seq_no;
    IPAddress curr_ip(curr->ip);
    struct far_entry old_fe; // valid route iff !is_new_rte
    bool is_new_rte = !rt_contains(curr_ip);
    if (is_new_rte)
      new_rtes.push_back(curr_ip);
    else
      old_fe = rt_get_ref(curr_ip);

    // insert this newly received route into the table if:
    if (is_new_rte                           // we don't already know about this nbr
	|| curr_seq > old_fe.nbr.seq_no      // or if recv'd seq_no is newer
	|| (curr_seq == old_fe.nbr.seq_no && // or if recv'd seq_no is same and hops are less
	    (curr->num_hops + 1) < old_fe.nbr.num_hops)) {
      far_entry fe(jiff, grid_nbr_entry(curr->ip, gh->ip, curr->num_hops + 1, curr->seq_no)); 
      fe.nbr.loc = curr->loc;
      fe.nbr.loc_err = curr->loc_err;
      fe.nbr.loc_good = curr->loc_good;
      if (curr->max_hops > 0)
	fe.nbr.max_hops = curr->max_hops - 1;
      fe.nbr.age = decr_age(curr->age, grid_hello::MIN_AGE_DECREMENT);
      rt_insert(fe);

      // if new entry is more than one hop away, remove from nbrs table also
      if (!is_new_rte && old_fe.nbr.num_hops > 1)
	_addresses.remove(curr_ip); // may fail, e.g. wasn't in nbr addresses table anyway


      if (_verbose)
	_agent->trace("S %.9f _%d_ %s route to %s, %d hops, next hop %s", 
		      jiff, _agent->ip_, is_new_rte ? "adding" : "updating", 
		      curr_ip.s().cc(), curr->num_hops + 1, IPAddress(gh->ip).s().cc());
    }
  }
    
  // remove the broken routes
  for (int i = 0; i < rtes_to_delete.size(); i++)
    rt_erase(rtes_to_delete[i]);

  // remember the latest triggered route info for transmission by the
  // paced trigger timer
  merge_triggered_routes(triggered_rtes, broken_rtes, new_rtes);

  if (do_proxy_stuff())
    pick_new_loc_proxy();
}


void
GridRT::merge_triggered_routes(Vector<far_entry> &good_routes, Vector<far_entry> &broken_routes,
			       Vector<IPAddress> &added_routes)
{
  // erase any outstanding broken route ads for which we now have a route
  for (int i = 0; i < added_routes.size(); i++) {
    int i2 = find_ip(_triggered_broken_routes, added_routes[i]);
    if (i2 >= 0) 
      _triggered_broken_routes.erase(i2);
  }

  // update any triggered good route ads
  for (int i = 0; i < good_routes.size(); i++) {
    int i2 = find_ip(_triggered_good_routes, good_routes[i].nbr.ip);
    if (i2 < 0)
      _triggered_good_routes.push_back(good_routes[i]);
    else {
      assert(good_routes[i].nbr.seq_no >= _triggered_good_routes[i2].nbr.seq_no); // or else we wouldn't have triggered it!
      _triggered_good_routes[i2] = good_routes[i];
    }
  }

  // remove any old incorrect good route ads, and merge in new broken route ads
  for (int i = 0; i < broken_routes.size(); i++) {
    int i2 = find_ip(_triggered_good_routes, broken_routes[i].nbr.ip);
    if (i2 >= 0)
      _triggered_good_routes.erase(i2);
    
    i2 = find_ip(_triggered_broken_routes, broken_routes[i].nbr.ip);
    if (i2 < 0)
      _triggered_broken_routes.push_back(broken_routes[i]);
    else {
      // can't trigger a broken route a second time without first
      // reinstating it as a good route.  that is, a route can only be
      // triggered as broken if it was first a good route.  but any
      // added new routes should cause outstanding triggered broken
      // routes advertisements to be thrown away first.

      assert(0); 
    }
  }
}


void
GridRT::process_grid_packet(Packet *p, grid_hdr *gh, double jiff)
{
  IPAddress ipaddr(gh->tx_ip);

  unsigned char ethbuf[6];
  int src_mac = _agent->mac_->hdr_src(p->access(_agent->off_cmn_));
  memcpy(ethbuf, &src_mac, 4); 
  EtherAddress ethaddr(ethbuf); 
  
  if (ethaddr == _ethaddr) {
    if (_verbose)
      printf("GridRT: received own Grid packet; ignoring it\n");
    return;
  }

  NbrEntry *nbr = _addresses.findp(ipaddr);
  if (nbr == 0) {
    // this src addr not already in map, so add it
    NbrEntry new_nbr(ethaddr, ipaddr, jiff);
    _addresses.insert(ipaddr, new_nbr);
  }
  else {
    // update jiffies and MAC for existing entry
    NbrEntry new_nbr = _addresses[ipaddr];
    new_nbr.last_updated_jiffies = jiff;
    if (new_nbr.eth != ethaddr) 
      new_nbr.eth = ethaddr;
    _addresses.insert(ipaddr, new_nbr);
  }
  
  /* XXX need to update (or add) routing entry in _rtes to be the
     correct one-hop entry for destination of this sender, ipaddr. */
}


void
GridRT::send_routing_update(Vector<grid_nbr_entry> &rte_info)
{
  assert((int) rte_info.size() <= _max_hello_entries);

  Packet *p = _agent->alloc_pkt(rte_info.size() * sizeof(grid_nbr_entry));
  struct hdr_cmn *cmnh = (struct hdr_cmn *) p->access(_agent->off_cmn_);
  struct hdr_ip *iph = (struct hdr_ip *) p->access(_agent->off_ip_);
  struct hdr_grid_lr *lrh = (struct hdr_grid_lr *) p->access(_agent->off_grid_lr_);
  struct grid_hdr *gh = &lrh->gh_;
  struct grid_hello *hlo = &lrh->hello_;

  // eth _ grid + hello + route entries
  cmnh->size_ = 14 + sizeof(grid_hdr) + sizeof(grid_hello) + rte_info.size()*sizeof(grid_nbr_entry);
  cmnh->ptype_ = PT_GRID_LR;
  cmnh->next_hop_ = MAC_BROADCAST;
  cmnh->addr_type_ = AF_LINK;

  cmnh->xmit_failure_ = 0; // can a broadcast fail?  we wouldn't care anyway.

  _agent->mac_->hdr_dst(p->access(_agent->off_mac_), MAC_BROADCAST);
  _agent->mac_->hdr_src(p->access(_agent->off_mac_), _agent->ip_);

  iph->dst_ = IP_BROADCAST;
  iph->dport_ = RT_PORT;
  assert((unsigned int) _agent->ip_ == ntohl(_ipaddr));
  iph->src_ = _agent->ip_;
  iph->sport_ = RT_PORT;

  gh->ip = _ipaddr;
  gh->tx_ip = _ipaddr;
  _agent->set_src_loc(gh);
  _agent->set_tx_loc(gh);

  assert(rte_info.size() <= 255);
  hlo->num_nbrs = (unsigned char) rte_info.size();
  hlo->seq_no = _seq_no;
  
  hlo->age = grid_hello::MAX_AGE_DEFAULT;

  if (do_proxy_stuff() && have_proxy()) 
    hlo->max_hops = rt_get(proxy_ip()).nbr.num_hops; // or add 1?

  grid_nbr_entry *curr = (grid_nbr_entry *) p->accessdata();
  for (int i = 0; i < rte_info.size(); i++) {
    *curr = rte_info[i];
    curr++;
  }

  Scheduler::instance().schedule(_agent->ifq_, p, 0);
}


bool 
GridRT::get_next_hop(IPAddress ip, IPAddress *next_hop)
{
  if (!rt_contains(ip))
    return false;
  *next_hop = rt_get_ref(ip).nbr.next_hop_ip;
  return true;
}


bool 
GridRT::get_next_geo_hop(IPAddress ip, IPAddress *next_hop)
{
  grid_location dst_loc;
  double last_adverised_jiff;
  bool found_dest_loc = get_advertised_loc(&dst_loc, &last_adverised_jiff, ntohl(ip));
  
  if (!found_dest_loc)
    return false;

  // only initialize outside of the loop if we actually know our
  // location...else initialize when we reach first route entry to a
  // node with location in the loop.
  double min_dist;
  if (know_loc()) {
    min_dist = grid_location::dist(dst_loc, get_loc());
  }

  // set here, if no nbrs are any good, next_hop will never get
  // changed to another addr
  *next_hop = _ipaddr; 

  // pick the node we have a route to that knows its location and is
  // closest to the ultimate destination
  bool first_with_loc = !know_loc(); // initialize in loop?
  for (FarTable::Iterator iter = _rtes.first(); iter; iter++) {
    /* we don't just check our neighbours, we check for any node N we
       have a route to --- each node along that route should know how
       to get to at least N, so will make a forwarding choice that has
       no worse progress than reaching N */
    if (iter.value().nbr.loc_good) {
      double d = grid_location::dist(dst_loc, iter.value().nbr.loc);
      if (d < min_dist || first_with_loc) {
	first_with_loc = false;
	min_dist = d;
	*next_hop = iter.value().nbr.next_hop_ip;
      }
    }
  }
  if (*next_hop != _ipaddr)
    return true;
  else
    return false;
}


void
GridRT::bad_next_hop(IPAddress bad_ip)
{
  // remove from nbr table
  _addresses.remove(bad_ip);

  // find routes with broken next hop
  Vector<far_entry> broken_routes;
  for (FarTable::Iterator i = _rtes.first(); i; i++) {
    if (IPAddress(i.value().nbr.next_hop_ip) == bad_ip)
      broken_routes.push_back(i.value());
  }

  // remove routes with broken next hop from table, and set them up to
  // merge with pending triggered routes.  I am skipping complicated
  // manipulation of ages here.  this whole code isn't very well
  // organized in that respect.
  for (int i = 0; i < broken_routes.size(); i++) {
    far_entry &f = broken_routes[i];
    rt_erase(f.nbr.ip);
    f.nbr.num_hops = 0;
    f.nbr.seq_no++;
    f.nbr.age = grid_hello::MAX_AGE_DEFAULT;
    f.when_triggered = Scheduler::instance().clock(); // now
  }  

  if (do_proxy_stuff())
    pick_new_loc_proxy();

  Vector<far_entry> t1(0);
  Vector<IPAddress> t2(0);
  merge_triggered_routes(t1, broken_routes, t2);
}

void
GridRT::dump()
{
  double jiff = Scheduler::instance().clock();
  printf("Routing table for %s\n", _ipaddr.s().cc());
  for (FarTable::Iterator iter = _rtes.first(); iter; iter++) {
    printf("ip=%s next_hop=%s num_hops=%u loc=%s loc_err=%hu %s seq_no=%u age=%f last_update=%f\n",
	   IPAddress(iter.value().nbr.ip).s().cc(),
	   IPAddress(iter.value().nbr.next_hop_ip).s().cc(),
	   (unsigned int) iter.value().nbr.num_hops,
	   iter.value().nbr.loc.s().cc(),
	   iter.value().nbr.loc_err,
	   iter.value().nbr.loc_good ? "loc-good" : "loc-bad",
	   iter.value().nbr.seq_no,
	   decr_age(iter.value().nbr.age, jiff - iter.value().last_updated_jiffies),
	   iter.value().last_updated_jiffies);
    assert(iter.value().nbr.num_hops > 0);
    assert(iter.value().nbr.ip > 0);
  }
  printf("End routing table for %s\n", _ipaddr.s().cc());
}


void
GridRT::rt_insert(far_entry fe)
{
  assert(fe.nbr.ip > 0); 
  assert(fe.nbr.num_hops > 0);
  _rtes.insert(fe.nbr.ip, fe);
  int sz = _rtes.size();
  GridAgent::max_rt_sz_ = max(GridAgent::max_rt_sz_, sz);
  if (do_proxy_stuff()) 
    GridAgent::max_ignorant_rt_sz_ = max(GridAgent::max_ignorant_rt_sz_, sz);
  if (do_proxy_stuff_with_loc())
    GridAgent::max_geo_rt_sz_ = max(GridAgent::max_geo_rt_sz_, sz);
}


struct GridRT::far_entry &
GridRT::rt_get_ref(IPAddress ip)
{
  assert(rt_contains(ip));
  return *_rtes.findp(ip);
}

struct GridRT::far_entry 
GridRT::rt_get(IPAddress ip)
{
  return rt_get_ref(ip);
}

void
GridRT::rt_erase(IPAddress ip)
{
  assert(rt_contains(ip));
  _rtes.remove(ip);
}


bool
GridRT::get_strawman_proxy_loc(grid_location *loc)
{
  int min_hops = -1;
  for (FarTable::Iterator iter = _rtes.first(); iter; iter++) {
    if (iter.value().nbr.loc_good && (iter.value().nbr.num_hops < min_hops || min_hops == -1)) {
      min_hops = iter.value().nbr.num_hops;
      *loc = iter.value().nbr.loc;
    }
  }
  if (min_hops > 0)
    return true;
  else
    return false;
}


int
GridRT::find_ip(Vector<grid_nbr_entry> v, IPAddress ip)
{
  for (int i = 0; i < v.size(); i++)
    if (v[i].ip == (unsigned int) ip)
      return i;
  return -1;
}

int
GridRT::find_ip(Vector<far_entry> v, IPAddress ip)
{
  for (int i = 0; i < v.size(); i++)
    if (v[i].nbr.ip == (unsigned int) ip)
      return i;
  return -1;
}

void
GridRT::swap(far_entry &f1, far_entry &f2)
{
  far_entry t = f1;
  f1 = f2;
  f2 = t;
}

int
GridRT::partition(Vector<far_entry> v, int p, int r)
{
  far_entry x = v[p];
  int i = p - 1;
  int j = r + 1;
  while (true) {
    do { j--; }
    while (!(v[j] <= x));
    do { i++; }
    while (!(x <= v[i]));
    if (i < j) 
      swap(v[i], v[j]);
    else
      return j;
  }      
}

void
GridRT::qsort(Vector<far_entry> v, int p, int r)
{
  // quicksort, CLR pg. 154
  if (p < r) {
    int q = partition(v, p, r);
    qsort(v, p, q);
    qsort(v, q + 1, r);
  }
}
