
#include "gfrt.hh"
#include "spanagent.hh"

#define TRACE_ROUTE    1

static inline double
compute_distance(double x1, double y1, double x2, double y2)
{
  return sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
}

static inline double
distance_between(int a, int b)
{
  double x, y, z, x2, y2, z2;
  (God::instance()->nodelist[a])->getLoc(&x, &y, &z);
  (God::instance()->nodelist[b])->getLoc(&x2, &y2, &z2);
  return compute_distance(x, y, x2, y2);
}

GFRouter::GFRouter(SPANAgent &a, SPANNeighborTable &t)
  : _agent(a), _neighbors(t)
{
}

int
GFRouter::find_next_hop(nsaddr_t dst)
{
#if TRACE_ROUTE > 1
  double now = Scheduler::instance().clock();
#endif

  //
  // if we are closest to the host, use us instead. we need this rule: if we
  // just withdraw, someone else might want to use us as a coordinator. if we
  // knew about our withdraw before that node (e.g. first withdraw message was
  // dropped due to collision), then without this rule, we may want to send
  // the packet back to that node, causing a loop.
  //
  double shortest_dist = distance_between(_agent.id(), dst);
  int hop = _agent.id();

  //
  // if we can forward the packet to destination, do it
  //
  if (shortest_dist < SPAN_RADIO_RANGE)
    return hop;

  //
  // first, look in neighbor table to get a coordinator closest to
  // the destination. we also use recently withdrawn coordinators.
  //
  SPANCoordinatorListIterator *ci = 
    _neighbors.coordinator_iterator(_agent.handle());
  while ((bool)(*ci)) { 
    //
    // remove any stale entries
    //
    if (distance_between(_agent.id(),ci->coordinator_id()) > SPAN_RADIO_RANGE){
      _neighbors.remove_coordinator(_agent.handle(), ci->coordinator_id()); 
      _neighbors.remove_neighbor(ci->coordinator_id());
    }
    else {
      double d = distance_between(ci->coordinator_id(), dst); 
#if TRACE_ROUTE > 1
      fprintf(stderr,"# %f %d: ->%d, %d looked at %f (%d)\n", 
	      now, _agent.id(), dst, ci->coordinator_id(), d, ci->withdrawn());
#endif
      if (d < shortest_dist || shortest_dist < 0) { 
#if TRACE_ROUTE > 1
        fprintf(stderr,"# %f %d: ->%d, %d shorested at %f\n", 
	        now, _agent.id(), dst, ci->coordinator_id(), d);
#endif
        shortest_dist = d; 
        hop = ci->coordinator_id(); 
      } 
    }
    ci->next();
  }
  delete ci;

  //
  // if we didn't find a route, searching through second hop coordinators. if
  // we find a second hop coordinator that is closer to the destination then
  // we are, we can send packets that way.
  //
  if (hop == _agent.id() && shortest_dist > SPAN_RADIO_RANGE) {
    SPANCoordinatorListIterator *ci = 
      _neighbors.coordinator_iterator(_agent.handle());
    while ((bool)(*ci)) { 
      int ch = _neighbors.search_neighbor(ci->coordinator_id());
      if (ch >= 0) {
        SPANCoordinatorListIterator *c2 = _neighbors.coordinator_iterator(ch);
        while ((bool)(*c2)) { 
	  if (c2->coordinator_id() != _agent.id()) { // prevent loops
            double d = distance_between(c2->coordinator_id(), dst);
#if TRACE_ROUTE > 1
            fprintf(stderr,"# %f %d: ->%d, %d looked at %f (via %d (%d))\n", 
	            now, _agent.id(), dst, c2->coordinator_id(), d, 
	            ci->coordinator_id(), ci->withdrawn());
#endif
            if (d < shortest_dist || shortest_dist < 0) {
              shortest_dist = d;
              hop = ci->coordinator_id();
            } 
	  }
	  c2->next(); 
	}
        delete c2;
      }
      ci->next();
    }
    delete ci;
  }

  return hop;
}

void
GFRouter::forward_pkt(Packet *p, int ip_off, int cmn_off, int span_off)
{
  hdr_ip *iph =  (hdr_ip*)p->access(ip_off); 
  hdr_cmn *cmnh =  (hdr_cmn*)p->access(cmn_off);

  if (--iph->ttl_ == 0) {
    _agent.drop(p, SPAN_GF_TTL);
    return;
  }

  if (iph->dst() == (int)IP_BROADCAST) {
    cmnh->next_hop_ = MAC_BROADCAST;
    _agent.send_packet(p);
    return;
  }

  int via = find_next_hop(iph->dst());
  assert(via >= 0);
  if (via == _agent.id() && 
      distance_between(_agent.id(), iph->dst()) > SPAN_RADIO_RANGE) {
#if TRACE_ROUTE
    fprintf(stderr,"# routing error %f: %d no route to near host %d\n", 
	    Scheduler::instance().clock(), _agent.id(), iph->dst());
#endif
    _neighbors.remove_neighbor(iph->dst());
    _agent.drop(p, SPAN_GF_HOLE);
  } else {
    if (via != _agent.id()) 
      cmnh->next_hop_ = via;
    else 
      cmnh->next_hop_ = iph->dst(); 
    _agent.send_packet(p); 
  }
}

