
#include "hdr_span.h"
#include "spanagent.hh"
#include "gfrt.hh"

#define StaticAssert(c) switch (c) case 0: case (c):
#define TRACE_ANNOUNCE   1
#define SPAN_FAIRNESS    1
#define ALL_COORDINATORS 0

// returns a random number between 0 and max, implemented by dmaltz
static inline double
jitter (double max, int be_random_)
{ 
  return (be_random_ ? Random::uniform(max) : 0);
}


//
// driver
//

static class SPANAgentClass : public TclClass
{
public:
  SPANAgentClass() : TclClass("Agent/SPANAgent") { }
  TclObject* create(int, const char *const *) { return new SPANAgent(); }
} class_SPANAgent;


SPANAgent::SPANAgent()
  : Agent(PT_SPAN),
    _periodic_broadcast_handler(this), 
    _check_announce_handler(this), 
    _check_withdraw_handler(this),
    _announce_coordinator_handler(this)
{
  _pkts_sent = 0;
  _pkts_received = 0;
  _pkts_hops = 0;

  bind("off_span_", &_off_span);
  bind("off_ll_", &_off_ll);
  bind("off_mac_", &_off_mac);
  bind("off_ip_", &_off_ip);

  _gfrt = new GFRouter(*this, _neighbors);
}

int
SPANAgent::command(int argc, const char*const* argv)
{
  TclObject *obj;  

  if (argc == 2) { 
    if (strcasecmp(argv[1], "start-span") == 0) { 
      start(); 
      return TCL_OK; 
    }
    if (strcasecmp(argv[1], "dump-tables") == 0) { 
      dump_table(); 
      return TCL_OK; 
    } 
  }
  else if(argc == 3) { 
    if (strcasecmp(argv[1], "ip-addr") == 0) { 
      _my_id = atoi(argv[2]); 
      return TCL_OK; 
    } 

    if (strcasecmp(argv[1],"drained-energy") == 0) {
      int energy = atoi(argv[2]); 
      _energy_meter += energy;
      return TCL_OK; 
    }

    if (strcasecmp(argv[1], "test-gf") == 0) { 
      nsaddr_t dst= atoi(argv[2]); 
      test_gf(dst); 
      return TCL_OK; 
    }
    if (strcasecmp(argv[1], "log-target") == 0) { 
      _logtarget = (Trace*) obj; 
      return TCL_OK; 
    }
    if (strcasecmp(argv[1],"add-real-ll") == 0 ) { 
      real_ll = (LL *) obj; 
      return TCL_OK; 
    } 
    else if (strcasecmp(argv[1], "mobile-node") == 0) {
      _mobile_node = (MobileNode *) TclObject::lookup(argv[2]);;
      if (_mobile_node == 0) {
        fprintf(stderr, "SPANAgent: %s lookup of %s failed\n", 
	        argv[1], argv[2]);
        return TCL_ERROR;
      }
      return TCL_OK;
    }
  }
  else if (argc == 4) {
    if (strcasecmp(argv[1], "add-mac") == 0) {
      _ifq = (PriQueue *) TclObject::lookup(argv[2]);
      if (_ifq == 0) {
        fprintf(stderr, "SPANAgent: %s lookup of %s failed\n", 
	        argv[1], argv[2]);
        return TCL_ERROR;
      }
      _mac = (Mac802_11 *) TclObject::lookup(argv[3]);
      if (_mac == 0) {
        fprintf(stderr, "SPANAgent: %s lookup of %s failed\n", 
	        argv[1], argv[3]);
        return TCL_ERROR;
      }
      return TCL_OK;
    }

    if (strcasecmp(argv[1], "add-ll") == 0) { 
      if( (obj = TclObject::lookup(argv[2])) == 0) { 
	fprintf(stderr, "SPANAgent: %s lookup of %s failed\n", 
	        argv[1], argv[2]); 
	return TCL_ERROR; 
      }
      _ll = (NsObject*) obj;
      if((obj = TclObject::lookup(argv[3])) == 0) { 
	fprintf(stderr, "SPANAgent: %s lookup of %s failed\n", 
	        argv[1], argv[3]); 
	return TCL_ERROR; 
      } 
      _ifq = (PriQueue *) obj; 
      return TCL_OK; 
    } 
  }

  return Agent::command(argc, argv);
}

void
SPANAgent::start()
{
  _is_coordinator = false;
  _announce_pending = false;
  _energy_meter = 0;
  _withdraw_meter = 0;
  _energy_trigger = 0;
  _last_withdraw = 0;
  _asleep = true;

  _neighbors.update_neighbor(_my_id, 0, 0, _is_coordinator);
  _self_handle = _neighbors.search_neighbor(_my_id);
  assert(_self_handle >=0 );
  double j = jitter(10,1);
  Scheduler::instance().schedule 
    (&_periodic_broadcast_handler, &_periodic_broadcast_event, j);
  j = 2+jitter(7, 1);
  Scheduler::instance().schedule 
    (&_check_announce_handler, &_check_announce_event, j);
}

void 
SPANAgent::recv(Packet *p, Handler *callback)
{ 
  hdr_ip *iph = (hdr_ip *) p->access (_off_ip); 
  hdr_cmn *cmnh =  (hdr_cmn*)p->access(off_cmn_);
  hdr_span *spanh =  (hdr_span*)p->access(_off_span);
  spanh->tries = 0;

  // deal with packet originated from us
  if (_my_id == iph->src()) {
    if (cmnh->num_forwards() == 0) {
      iph->ttl_ = SPAN_DEF_TTL;
      spanh->type = SPAN_ROUTE_GF;
      cmnh->size() += IP_HDR_LEN + spanh->piggyback_size();
    }
    _gfrt->forward_pkt(p, _off_ip, off_cmn_, _off_span); 
    return;
  }
  
  if (iph->dst() == _my_id) {
    _pkts_received++;
    _pkts_hops += ((SPAN_DEF_TTL-iph->ttl())+1);
    process_span_hdr(p);
    drop(p, SPAN_DROP_RECEIVED);
    return;
  }

  // assume link layer will buffer broadcast messages
  if (_asleep && cmnh->next_hop_ != (int)MAC_BROADCAST &&
      (Scheduler::instance().clock()-_last_withdraw > SPAN_WITHDRAW_GRACE)) {
    if (spanh->type == SPAN_ROUTE_GF)
      drop(p, SPAN_DROP_COORD);
    else 
      drop(p, SPAN_DROP_SLEEP);
    return;
  }

  if (spanh->type == SPAN_ROUTE_GF || spanh->type == SPAN_HELLO) 
    process_span_hdr(p);

  if (spanh->type == SPAN_ROUTE_GF) {
    _gfrt->forward_pkt(p, _off_ip, off_cmn_, _off_span); 
    return;
  } 
  
  Packet::free(p);
}

void
SPANAgent::sleep()
{
  assert(!_is_coordinator);
  _asleep = true;
  double d = Scheduler::instance().clock() - _energy_trigger;
  _energy_meter += d;
  // wait longer before we check again, give someother node a chance 
  // to become coordinator
  Scheduler::instance().schedule
    (&_check_announce_handler, &_check_announce_event, 2*SPAN_ANNOUNCE_CHECK);
}

void
SPANAgent::wakeup()
{
  _energy_trigger = Scheduler::instance().clock();
  _asleep = false;
}

void 
SPANAgent::process_span_hdr(Packet *p)
{
  hdr_span *spanh =  (hdr_span*)p->access(_off_span);
 
  int handle;
  // add/update node to neighbor table
  if ((handle = _neighbors.update_neighbor 
	(spanh->node_id, 1, spanh->nounce, spanh->is_coordinator)) < 0) 
    return;
 
  // update coordinator information
  if (spanh->is_coordinator)
    _neighbors.add_coordinator(_self_handle, spanh->node_id);
  else {
    if (spanh->is_asleep)
      _neighbors.remove_coordinator(_self_handle, spanh->node_id);
    else
      _neighbors.withdrawn_coordinator(_self_handle, spanh->node_id);
  }

  if (spanh->type == SPAN_HELLO) {
    
    // save coordinators of that node
    _neighbors.clear_coordinators(handle);
    for (unsigned i=0; i<spanh->ncoordinators; i++)
      _neighbors.add_coordinator(handle, spanh->coordinators[i]);
   
    // add second hop neighbors for that node
    _neighbors.clear_second_hops(handle);
    for (unsigned i=0; i<spanh->nneighbors; i++)
      _neighbors.add_second_hops(handle, spanh->neighbors[i]);
  
    // be nice, update other neighbor's coordinator list:
    // this helps propagation of topology...
  
    SPANNeighborTableIterator *iter = _neighbors.iterator();
    while ((bool)(*iter)) {
      if (iter->neighbor_id() != _my_id) {
        int nh = iter->handle();
        _neighbors.remove_coordinator(nh, spanh->node_id);
      }
      iter->next();
    }
    delete iter;
  
    for(unsigned i=0; i<spanh->nneighbors; i++) {
      int node = spanh->neighbors[i];
      int nh;
      if ((nh = _neighbors.search_neighbor(node)) >= 0) {
        if (spanh->is_coordinator)
          _neighbors.add_coordinator(nh, spanh->node_id);
      }
    }
  }
}


//
// announce and withdraw
//

int
SPANAgent::check_announce_coordinator()
{ 
#if ALL_COORDINATORS
  return 100;
#else
  if (_is_coordinator) return 0;
  int n = 0;

  SPANNeighborTableIterator *iter = _neighbors.iterator(); 
  while((bool)(*iter)) { 
    int node = iter->neighbor_id();
    SPANNeighborTableIterator *iter2 = _neighbors.iterator(); 
    while((bool)(*iter2)) {
      if (iter2->neighbor_id() > node) {
	if (_neighbors.share_coordinator(iter2->neighbor_id(),node) < 0)
	  n++;
      }
      iter2->next();
    }
    delete iter2;
    iter->next(); 
  }
  delete iter;
  return n;
#endif
}

bool
SPANAgent::check_withdraw_coordinator()
{
#if ALL_COORDINATORS
  return false;
#else
#if SPAN_FAIRNESS
  bool force_withdraw = false;
  _withdraw_meter++;
  if (_withdraw_meter == 10) {
    _withdraw_meter = 0;
    force_withdraw = true;
  }
#endif

  bool withdraw = true;
  SPANNeighborTableIterator *iter = _neighbors.iterator(); 
 
  // check each neighbor. do not withdraw if two
  // neighbors can't reach each other without me
  
  while((bool)(*iter) && withdraw) {
    if (iter->neighbor_id() == _my_id) {
      iter->next();
      continue;
    }
    if (withdraw) {
      SPANNeighborTableIterator *iter2 = _neighbors.iterator(); 
      while((bool)(*iter2) && withdraw) {
        if (iter2->neighbor_id() == iter->neighbor_id() || 
	    iter2->neighbor_id() == _my_id) {
	  iter2->next();
	  continue;
        }
	int joint;
	if ((joint = _neighbors.share_coordinator 
	      (iter->neighbor_id(), iter2->neighbor_id(), _my_id)) < 0) {
#if SPAN_FAIRNESS
	  if (force_withdraw) {
            if ((joint = _neighbors.share_neighbor
		  (iter->neighbor_id(), iter2->neighbor_id(), _my_id)) < 0) {
#if TRACE_ANNOUNCE > 1
	      fprintf(stderr, "# %d ck withdraw: %d %d can't connect\n",
		  _my_id, iter->neighbor_id(), iter2->neighbor_id());
#endif
	      withdraw = false;
	    }
	  }
	  else 
#endif
	  withdraw = false;
	}
        iter2->next();
      }
      delete iter2;
    }
    iter->next();
  }
  delete iter;
  return withdraw;
#endif
}

void
SPANCheckAnnounceHandler::handle(Event *e)
{
  if (_agent->_is_coordinator) return;
  if (!_agent->_announce_pending) { 
    // decide if we should become a coordinator 
    int c;
    if ((c = _agent->check_announce_coordinator())>0) { 
      double energy = 1-(_agent->_energy_meter/_agent->_max_energy_meter);
      // if we still have some energy left
      if (energy > 0.1) { 
	// calculate a delay based on n and 
	// energy_meter. more n means higher priority.
        double min_delay = (_agent->_neighbors.nneighbors())/(c*energy);
	double j = jitter(min_delay,1);
#if TRACE_ANNOUNCE > 2
	fprintf(stderr,"# %d: delay = %f (min %f, c %d, energy %f)\n", 
	        _agent->id(), j, min_delay, c, energy);
#endif
	_agent->_announce_pending = true;
        Event *ev = new Event; 
        Scheduler::instance().schedule 
	  (&_agent->_announce_coordinator_handler, ev, j);
      }
    }
  }
  Scheduler::instance().schedule(this, e, SPAN_ANNOUNCE_CHECK);
}

void
SPANAnnounceCoordinatorHandler::handle(Event *e)
{
  _agent->announce_coordinator();
  delete e;
}

void
SPANAgent::announce_coordinator()
{
  _announce_pending = false;

  if (check_announce_coordinator() > 0) {
    if (_asleep) wakeup();

#if TRACE_ANNOUNCE
    fprintf(stderr,"# %f: %d announce coordinator\n", 
	    Scheduler::instance().clock(), _my_id);
#endif
    _is_coordinator = true;

    _neighbors.update_neighbor(_my_id, 0, Scheduler::instance().clock(), true);
    _neighbors.add_coordinator(_self_handle, _my_id);
   
    broadcast_hello();
    Scheduler::instance().schedule
      (&_check_withdraw_handler, &_check_withdraw_event, SPAN_WITHDRAW_CHECK);
  } 
}

void
SPANCheckWithdrawHandler::handle(Event *e)
{
  if (_agent->is_coordinator()) {
    if (_agent->check_withdraw_coordinator()) {
      _agent->withdraw_coordinator(); 
      return;
    }
    Scheduler::instance().schedule(this,e,SPAN_WITHDRAW_CHECK);
  }
}

void
SPANAgent::withdraw_coordinator()
{
#if TRACE_ANNOUNCE
  fprintf(stderr,"# %f: %d withdraw coordinator (%d)\n", 
	  Scheduler::instance().clock(), _my_id, _asleep);
#endif
  _is_coordinator = false;
  _last_withdraw = Scheduler::instance().clock();
  _neighbors.update_neighbor(_my_id, 0, Scheduler::instance().clock(), false);
  _neighbors.remove_coordinator(_self_handle, _my_id);
  broadcast_hello();
}


//
// broadcast
//

void
SPANPeriodicBroadcastHandler::handle(Event *e)
{
  _agent->broadcast_hello();
  if (!_agent->is_coordinator() && !_agent->asleep()) {
    if (_agent->last_withdraw()+(SPAN_WITHDRAW_GRACE/2)
	<= Scheduler::instance().clock()) {
#if TRACE_ANNOUNCE > 1
      fprintf(stderr,"# %f: %d withdrawn, goto sleep\n", 
	      Scheduler::instance().clock(), _agent->id());
#endif
      _agent->sleep();
    }
  }
  Scheduler::instance().schedule(this,e,SPAN_BCAST);
}

void
SPANAgent::piggyback_span_hdr(Packet *p)
{
  Scheduler & s = Scheduler::instance();
  double now = s.clock();

  hdr_span *spanh =  (hdr_span*)p->access(_off_span);
  spanh->is_coordinator = _is_coordinator;
  spanh->is_asleep = _asleep;
  spanh->node_id = _my_id;
  spanh->nounce = now;
}

void
SPANAgent::broadcast_hello()
{
  Scheduler & s = Scheduler::instance();
  double now = s.clock();

  // first refresh our neighbors table
  _neighbors.expire_old_entries(_my_id, now);
  _neighbors.refresh_coordinators(_my_id);

  Packet *p = allocpkt();
  piggyback_span_hdr(p);
  
  hdr_cmn *cmnh =  (hdr_cmn*)p->access(off_cmn_);
  hdr_span *spanh =  (hdr_span*)p->access(_off_span);
  spanh->type = SPAN_HELLO;

  unsigned i = 0;
  SPANCoordinatorListIterator *citer = 
    _neighbors.coordinator_iterator(_self_handle);
  while ((bool)(*citer) && i < MAX_SPAN_COORDINATORS) {
    if (!citer->withdrawn()) {
      spanh->coordinators[i] = citer->coordinator_id(); 
      i++;
    }
    citer->next();
  }
  delete citer;
  spanh->ncoordinators = i;

  i = 0;
  SPANNeighborTableIterator *iter = _neighbors.iterator();
  while ((bool)(*iter) && i < MAX_SPAN_NEIGHBORS) {
    spanh->neighbors[i] = iter->neighbor_id(); 
    i++;
    iter->next();
  }
  delete iter;
  spanh->nneighbors = i;

  cmnh->next_hop_ = MAC_BROADCAST;
  // XXX - we don't really use the IP header: just need it to get ll to work
  // correctly in ns. thus, don't include IP header size;
  cmnh->size() = spanh->size();

  hdr_ip *iph =  (hdr_ip*)p->access(off_ip_);
  iph->dst() = IP_BROADCAST;
  iph->src() = _my_id;

  send_packet(p);
  _neighbors.update_neighbor
    (_my_id, 0, Scheduler::instance().clock(), _is_coordinator);
}

  
//
// transmitter
//

void
SPANAgent::send_packet(Packet *p)
{
  hdr_cmn *cmnh =  (hdr_cmn*)p->access(off_cmn_);
  hdr_span *spanh =  (hdr_span*)p->access(_off_span);
  if (spanh->type == SPAN_ROUTE_GF)
    piggyback_span_hdr(p);
  cmnh->addr_type_ = AF_INET;
  cmnh->xmit_failure_ = tx_failure_cb;
  cmnh->xmit_failure_data_ = (void*) this;
#if USE_ARP 
  double delay = spanh->delay;
  spanh->delay = 0;
  Scheduler::instance().schedule(_ll, p, delay);
#else
  _mac->hdr_dst(p->access(_off_mac), cmnh->next_hop_);
  _mac->hdr_src(p->access(_off_mac), _my_id);
  Scheduler::instance().schedule(_ifq, p, 0);
#endif
}

void
SPANAgent::tx_failure_cb(Packet *p, void *thunk)
{
  SPANAgent *agent = (SPANAgent*) thunk;
  hdr_span *spanh =  (hdr_span*)p->access(agent->_off_span);
  if (spanh->tries < 4) {
    spanh->delay = 0.01;
    spanh->tries++;
    agent->send_packet(p);
  }
  else
    agent->drop(p, SPAN_DROP_TX_FAILED);
}


//
// traffic generator
//

void
SPANAgent::test_gf(nsaddr_t dst)
{
  Event *ev = new Event; 
  SPANTestGFHandler *h = new SPANTestGFHandler(this, dst);
  double j = jitter(_send_interval, 1);
  Scheduler::instance().schedule(h, ev, j);
}

void
SPANTestGFHandler::handle(Event *e)
{
  if (_pkts < 16) {
    _agent->make_pkt(_dst, SPAN_ROUTE_GF);
    _pkts++;
    Scheduler::instance().schedule(this,e,_agent->_send_interval);
  } else {
    delete e;
    delete this;
  }
}

void
SPANAgent::make_pkt(nsaddr_t dst, unsigned type)
{
  Packet *p = allocpkt();
  hdr_ip *iph =  (hdr_ip*)p->access(_off_ip); 
  hdr_cmn *cmnh =  (hdr_cmn*)p->access(off_cmn_);
  hdr_span *spanh =  (hdr_span*)p->access(_off_span);
  
  cmnh->ptype_ = PT_SPAN;
  spanh->type = type;
  iph->dst() = dst;
  iph->dport_ = RT_PORT;
  iph->src() = _my_id;
  spanh->delay = 0;

  _pkts_sent++;
  cmnh->size() = spanh->size();
  recv(p);
}


//
// debugging
//

void
SPANAgent::dump_table()
{
  if (_my_id == 1) 
    fprintf(stderr,"StartOfDump\n");
  if (_is_coordinator) {
    _energy_meter += Scheduler::instance().clock() - _energy_trigger;
    _energy_trigger = Scheduler::instance().clock();
  }
  SPANNeighborTableIterator *iter = _neighbors.iterator();
  double x, y, z;
  (God::instance()->nodelist[_my_id])->getLoc(&x, &y, &z);
  while((bool)(*iter)) {
    fprintf(stderr, "%f %f %d %c%d %d %d %f %f %d %d %d\n",
	    x, y, _my_id, iter->is_coordinator() ? 'c' : 'n',
	    iter->neighbor_id(), iter->neighbor_id(),
	    iter->hops(), iter->neighbor_nounce(), 
	    _energy_meter, _pkts_sent, _pkts_received, _pkts_hops);
    iter->next();
  }
  delete iter;
}


