#ifndef SPANNEIGHBORTABLE_HH
#define SPANNEIGHBORTABLE_HH

#include <stdio.h>
#include <vector>

class SPANNeighborTable {
  static const int _expire_period = 11;

  friend class SPANNeighborTableIterator;
  friend class SPANCoordinatorListIterator;
  friend class GFRouter;

protected:
  struct SPANCoordinatorInfo {
    bool in_use;
    bool withdrawn;
    int coordinator_id;
  };
  struct SPANNeighborInfo {
    bool in_use;
    bool is_coordinator;
    int neighbor_id;
    int hops;
    double neighbor_nounce;
    vector<SPANCoordinatorInfo> coordinators;
    vector<int> second_hops;
  };
  vector<struct SPANNeighborInfo> _neighbors;

  // search for neighbor or coordinator information, if found return handle,
  // otherwise, return -1, and free is set to the first entry in vector that
  // is free and can be reused
  int search_neighbor(int node, int &free);
  int search_coordinator(int node_h, int coord, int &free);

  int share_coordinator(int a, int b, int ex, bool use_ex);
  int share_neighbor(int a, int b, int ex, bool use_ex);

public:
  SPANNeighborTable() { _neighbors.clear(); }
  SPANNeighborTableIterator* iterator() const;
  SPANNeighborTableIterator* iterator(int handle) const;
  SPANCoordinatorListIterator* coordinator_iterator(int handle) const;

  int nneighbors() const { 
    int n = 0;
    for(unsigned i=0; i< _neighbors.size(); i++) {
      if (_neighbors[i].in_use) n++;
    }
    return n;
  }

  int neighbor_id(int handle) const {
    assert(handle >= 0);
    assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
    return _neighbors[handle].neighbor_id;
  }

  double neighbor_nounce(int handle) const {
    assert(handle >= 0);
    assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
    return _neighbors[handle].neighbor_nounce;
  }

  int neighbor_ncoordinators(int handle) const {
    assert(handle >= 0);
    assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
    int n = 0;
    for (unsigned i=0; i<_neighbors[handle].coordinators.size(); i++)
      if (_neighbors[handle].coordinators[i].in_use &&
	  !_neighbors[handle].coordinators[i].withdrawn) n++;
    return n;
  }

  bool neighbor_is_coordinator(int handle) const {
    assert(handle >= 0);
    assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
    return _neighbors[handle].is_coordinator;
  }

  // updates neighbor information. returns handle to the neighbor if nounce is
  // newer, otherwise returns -1
  int update_neighbor(int node, int hops, double nounce, bool is_c);

  // expects first argument to be a neighbor handle returned by
  // update_neighbor or search_neighbor. adds coordinator information.
  // returns handle to the coordinator.
  int add_coordinator(int node_h, int c);

  // clears second hop neighbors list of a neighbor
  int clear_second_hops(int node_h);

  // add second hop neighbor to a neighbor
  int add_second_hops(int node_h, int node);

  // returns node or coordinator handle if node or coordinator exists,
  // otherwise, -1
  int search_neighbor(int node);
  int search_coordinator(int node_h, int coord);

  // removes a neighbor from table. returns 0 if successful, otherwise
  // (neighbor does not exist) returns -1
  int remove_neighbor(int node);

  // removes all coordinator information
  void clear_coordinators(int node_h);

  // expects first argument to be a neighbor handle. removes a coordinator
  // from that neighbor. returns 0 if successful, otherwise (coordinator does
  // not exist) returns -1
  int remove_coordinator(int node_h, int c);
  int withdrawn_coordinator(int node_h, int c);
 
  // returns coordinator that joins neighbors a and b. -1 if they 
  // aren't joined.
  int share_coordinator(int node_a, int node_b);

  // returns coordinator that joins neighbors a and b. -1 if they
  // aren't joined. don't consider the exclude node as a join. 
  int share_coordinator(int node_a, int node_b, int exclude);
  int share_neighbor(int node_a, int node_b, int exclude);

  // expire table entries at least "expiration" old. don't expire the
  // "exclude" node.
  int expire_old_entries(int exclude, double now);

  // update list of coordinators for the given node. if
  // a coordinator is no longer in neighbor table,
  // remove it from the list. this function should only
  // be called for the self node, since for other nodes
  // their coordinators may not be in range. return 0 if
  // successful. returns -1 otherwise.
  int refresh_coordinators(int node);
};
  
int inline
SPANNeighborTable::search_neighbor(int node, int& free)
{
  free = -1;
  for (unsigned i=0; i<_neighbors.size(); i++) {
    if (_neighbors[i].in_use && _neighbors[i].neighbor_id == node)
      return i;
    else if (!_neighbors[i].in_use && free < 0)
      free = i;
  }
  return -1;
}

int inline
SPANNeighborTable::search_neighbor(int node)
{
  int free;
  return search_neighbor(node, free);
}

int inline
SPANNeighborTable::remove_neighbor(int node)
{
  int i;
  if ((i = search_neighbor(node)) >= 0) {
    _neighbors[i].in_use = false;
    _neighbors[i].coordinators.clear();
    return 0;
  }
  return -1;
}
  
int inline
SPANNeighborTable::clear_second_hops(int handle)
{
  assert(handle >= 0);
  assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
  _neighbors[handle].second_hops.clear();
  return 0;
}

int inline
SPANNeighborTable::add_second_hops(int handle, int node)
{
  assert(handle >= 0);
  assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
  _neighbors[handle].second_hops.push_back(node);
  return 0;
}

int inline
SPANNeighborTable::search_coordinator(int node_h, int c, int &free)
{
  assert(node_h >= 0);
  assert((unsigned)node_h < _neighbors.size() && _neighbors[node_h].in_use);
  free = -1;
  for (unsigned j=0; j<_neighbors[node_h].coordinators.size(); j++) { 
    if (_neighbors[node_h].coordinators[j].in_use && 
	_neighbors[node_h].coordinators[j].coordinator_id == c)
      return j;
    else if (!_neighbors[node_h].coordinators[j].in_use) {
      if (free < 0) free = j;
      if (_neighbors[node_h].coordinators[j].coordinator_id == c)
	free = j;
    }
  }
  return -1;
}
  
int inline
SPANNeighborTable::search_coordinator(int node_h, int c)
{
  assert(node_h >= 0);
  assert((unsigned)node_h < _neighbors.size() && _neighbors[node_h].in_use);
  int free;
  return search_coordinator(node_h, c, free);
}
  
void inline
SPANNeighborTable::clear_coordinators(int node_h)
{
  assert(node_h >= 0);
  assert((unsigned)node_h < _neighbors.size() && _neighbors[node_h].in_use);
  _neighbors[node_h].coordinators.clear();
}

int inline
SPANNeighborTable::withdrawn_coordinator(int node_h, int c)
{
  assert(node_h >= 0);
  assert((unsigned)node_h < _neighbors.size() && _neighbors[node_h].in_use);
  int i;
  if ((i = search_coordinator(node_h, c)) >= 0) {
    _neighbors[node_h].coordinators[i].withdrawn = true;
    return 0;
  }
  return -1;
}
  
int inline
SPANNeighborTable::remove_coordinator(int node_h, int c)
{
  assert(node_h >= 0);
  assert((unsigned)node_h < _neighbors.size() && _neighbors[node_h].in_use);
  int i;
  if ((i = search_coordinator(node_h, c)) >= 0) {
    _neighbors[node_h].coordinators[i].in_use = false;
    _neighbors[node_h].coordinators[i].withdrawn = true;
    return 0;
  }
  return -1;
}
  
int inline
SPANNeighborTable::share_coordinator(int a, int b)
{
  return share_coordinator(a, b, 0, false);
}

int inline
SPANNeighborTable::share_coordinator(int a, int b, int ex)
{
  return share_coordinator(a, b, ex, true);
}

int inline
SPANNeighborTable::share_neighbor(int a, int b, int ex)
{
  return share_neighbor(a, b, ex, true);
}


class SPANNeighborTableIterator
{
protected:
  friend class SPANNeighborTable;
  
  const SPANNeighborTable *_table;
  int _handle;
  SPANNeighborTableIterator(const SPANNeighborTable *t, int h);

public:
  int neighbor_id() const;
  double neighbor_nounce() const;
  int hops() const;
  bool is_coordinator() const;
  int ncoordinators() const;

  void reset();
  void next();
  int handle() const	{ return _handle; }
  operator bool() const;
};
  

inline
SPANNeighborTableIterator::SPANNeighborTableIterator
  (const SPANNeighborTable *t, int h) : _table(t), _handle(h) {
  assert(_handle >= 0);
  if ((unsigned)_handle<_table->_neighbors.size() &&
      _table->_neighbors[_handle].in_use == false) next();
}

inline SPANNeighborTableIterator* 
SPANNeighborTable::iterator() const 
{ 
  return new SPANNeighborTableIterator(this, 0);
}

inline SPANNeighborTableIterator* 
SPANNeighborTable::iterator(int handle) const 
{ 
  return new SPANNeighborTableIterator(this, handle);
}

inline void
SPANNeighborTableIterator::reset()
{
  _handle = 0;
  if ((unsigned)_handle<_table->_neighbors.size() &&
      _table->_neighbors[_handle].in_use == false) 
    next();
}

inline void
SPANNeighborTableIterator::next()
{
  do {
    _handle++;
  }
  while ((unsigned)_handle<_table->_neighbors.size() &&
         _table->_neighbors[_handle].in_use == false);
}

inline
SPANNeighborTableIterator::operator bool() const
{
  return (unsigned)_handle < _table->_neighbors.size();
}

inline int
SPANNeighborTableIterator::neighbor_id() const
{
  assert((bool)(*this));
  return _table->_neighbors[_handle].neighbor_id;
}

inline double
SPANNeighborTableIterator::neighbor_nounce() const
{
  assert((bool)(*this));
  return _table->_neighbors[_handle].neighbor_nounce;
}

inline int
SPANNeighborTableIterator::hops() const
{
  assert((bool)(*this));
  return _table->_neighbors[_handle].hops;
}

inline bool
SPANNeighborTableIterator::is_coordinator() const
{
  assert((bool)(*this));
  return _table->_neighbors[_handle].is_coordinator;
}

inline int
SPANNeighborTableIterator::ncoordinators() const
{
  assert((bool)(*this));
  return _table->_neighbors[_handle].coordinators.size();
}


class SPANCoordinatorListIterator
{
protected:
  friend class SPANNeighborTable;
  
  const SPANNeighborTable *_table;
  int _nhandle;
  int _chandle;

  SPANCoordinatorListIterator(const SPANNeighborTable *t, int nh);

public:
  int coordinator_id() const;

  void reset();
  void next();
  int handle() const	{ return _chandle; }
  operator bool() const;
  bool in_use() const;
  bool withdrawn() const;
};
 
inline
SPANCoordinatorListIterator::SPANCoordinatorListIterator
  (const SPANNeighborTable *t, int nh) 
  : _table(t), _nhandle(nh), _chandle(0) 
{
  assert(_nhandle >= 0);
  if ((unsigned)_chandle < _table->_neighbors[_nhandle].coordinators.size() &&
      _table->_neighbors[_nhandle].coordinators[_chandle].in_use == false) 
    next();
}

inline SPANCoordinatorListIterator* 
SPANNeighborTable::coordinator_iterator(int handle) const
{ 
  assert(handle >= 0);
  assert((unsigned)handle < _neighbors.size() && _neighbors[handle].in_use);
  return new SPANCoordinatorListIterator(this, handle); 
}

inline void
SPANCoordinatorListIterator::reset()
{
  _chandle = 0;
  if ((unsigned)_chandle < _table->_neighbors[_nhandle].coordinators.size() &&
      _table->_neighbors[_nhandle].coordinators[_chandle].in_use == false) 
    next();
}

inline void
SPANCoordinatorListIterator::next()
{
  do {
    _chandle++;
  }
  while ((unsigned)_chandle<_table->_neighbors[_nhandle].coordinators.size() &&
         _table->_neighbors[_nhandle].coordinators[_chandle].in_use == false);
}

inline
SPANCoordinatorListIterator::operator bool() const
{
  return (unsigned)_chandle < _table->_neighbors[_nhandle].coordinators.size();
}

inline int
SPANCoordinatorListIterator::coordinator_id() const
{
  assert((bool)(*this));
  return _table->_neighbors[_nhandle].coordinators[_chandle].coordinator_id;
}

inline bool
SPANCoordinatorListIterator::in_use() const
{
  assert((bool)(*this));
  return _table->_neighbors[_nhandle].coordinators[_chandle].in_use;
}

inline bool
SPANCoordinatorListIterator::withdrawn() const
{
  assert((bool)(*this));
  return _table->_neighbors[_nhandle].coordinators[_chandle].withdrawn;
}

#endif


